@@ -1,123 +1,114 @@
+# git-crecord entry point and configuration helpers
+#
+# Copyright 2016, 2018—2022 Andrej Shadura <andrew@shadura.me>
+#
+# This software may be used and distributed according to the terms of
+# the GNU General Public License, incorporated herein by reference.
+#
+# SPDX-License-Identifier: GPL-2.0-or-later
+
 from gettext import gettext as _
-from .gitrepo import GitRepo
+from pathlib import Path
+from typing import Optional
+
+import argparse
 import os
 import sys
-from . import crecord_core
-from . import util
 import tempfile
-import argparse
+
+from . import crecord_core
+from .gitrepo import GitRepo
+from .util import Abort, system, systemcall
+
 
 class Config:
-    def get(self, section, item, default=None):
+    def get(self, section, item, default=None) -> Optional[str]:
         try:
-            return util.systemcall(['git', 'config', '--get', '%s.%s' % (section, item)], onerr=KeyError).rstrip('\n')
+            return systemcall(
+                ['git', 'config', '--get', '%s.%s' % (section, item)],
+                onerr=KeyError,
+                encoding="UTF-8",
+            ).rstrip('\n')
         except KeyError:
             return default
 
     def set(self, section, item, value, source=""):
         raise NotImplementedError
 
+
 class Ui:
-    def __init__(self, repo):
+    def __init__(self, repo: GitRepo):
         self.repo = repo
         self.config = Config()
         self.debuglevel = 0
-        try:
-            self._username = "%s <%s>" % (self.config.get("user", "name"), self.config.get("user", "email"))
-        except KeyError:
-            self._username = None
+
+    def print_message(self, *msg, debuglevel: int, **opts):
+        if self.debuglevel < debuglevel:
+            return
+
+        sys.stdout.flush()
+        print(*msg, **opts, file=sys.stderr)
+        sys.stderr.flush()
 
     def debug(self, *msg, **opts):
-        if self.debuglevel < 2:
-            return
-        for m in msg:
-            sys.stdout.write(m)
+        self.print_message(*msg, debuglevel=2, **opts)
 
     def info(self, *msg, **opts):
-        if self.debuglevel < 1:
-            return
-        sys.stdout.flush()
-        for m in msg:
-            sys.stderr.write(m)
-        sys.stderr.flush()
-
-    def status(self, *msg, **opts):
-        for m in msg:
-            sys.stdout.write(m)
+        self.print_message(*msg, debuglevel=1, **opts)
 
     def warn(self, *msg, **opts):
-        sys.stdout.flush()
-        for m in msg:
-            sys.stderr.write(m)
-        sys.stderr.flush()
+        self.print_message(*msg, debuglevel=0, **opts)
+
+    def status(self, *msg, **opts):
+        print(*msg, **opts)
 
     def setdebuglevel(self, level):
         self.debuglevel = level
 
-    def setusername(self, username):
-        self._username = username
-
-    def username(self):
-        if self._username is None:
-            util.Abort(_("no name or email for the author was given"))
-        return self._username
-
-    def geteditor(self):
-        editor = 'sensible-editor'
+    @property
+    def editor(self) -> str:
         return (os.environ.get("GIT_EDITOR") or
                 self.config.get("core", "editor") or
                 os.environ.get("VISUAL") or
-                os.environ.get("EDITOR", editor))
+                os.environ.get("EDITOR") or
+                'sensible-editor')
 
-    def edit(self, text, user, extra=None, name=None):
-        fd = None
-        if name is None:
-            (fd, name) = tempfile.mkstemp(prefix='git-crecord-',
-                                      suffix=".txt", text=True)
+    def edit(self, text: bytes, user, extra=None, name=None) -> bytes:
+        f = tempfile.NamedTemporaryFile(
+            prefix='git-crecord-',
+            suffix=".txt",
+            mode="wb",
+            delete=False,
+        )
         try:
-            if fd is not None:
-                f = os.fdopen(fd, "w")
-            else:
-                f = open(name, "w")
             f.write(text)
             f.close()
 
-            editor = self.geteditor()
+            editor = self.editor
 
-            util.system("%s \"%s\"" % (editor, name),
-                       onerr=util.Abort, errprefix=_("edit failed"))
+            system("%s \"%s\"" % (editor, f.name),
+                   onerr=Abort, errprefix=_("edit failed"))
+
+            t = Path(f.name).read_bytes()
 
-            f = open(name)
-            t = f.read()
-            f.close()
         finally:
-            if fd is not None:
-                os.unlink(name)
+            os.unlink(f.name)
 
         return t
 
     def stage(self, *files, **opts):
         to_add = [f for f in files if os.path.exists(f)]
         if to_add:
-            util.system(['git', 'add', '-f', '-N', '--'] + to_add,
-                       onerr=util.Abort, errprefix=_("add failed"))
+            system(['git', 'add', '-f', '-N', '--'] + to_add,
+                   onerr=Abort, errprefix=_("add failed"))
 
     def commit(self, *files, **opts):
-        (fd, name) = tempfile.mkstemp(prefix='git-crecord-',
-                                      suffix=".txt", text=True)
+        msgfile = self.repo.controldir / "CRECORD_COMMITMSG"
         try:
             args = []
 
-            # git-commit doesn't play nice with empty lines
-            # and comments in the commit message when --edit
-            # is used with --file;
-            # to work that around, use --template when
-            # no message is specified and --file otherwise.
-
-            f = os.fdopen(fd, "w")
-            if opts['message'] is not None:
-                f.write(opts['message'])
-            f.close()
+            if opts['message']:
+                msgfile.write_text(opts['message'])
 
             if opts['cleanup'] is None:
                 opts['cleanup'] = 'strip'
@@ -125,7 +116,7 @@ class Ui:
             for k, v in opts.items():
                 if k in ('author', 'date', 'amend', 'signoff', 'cleanup',
                          'reset_author', 'gpg_sign', 'no_gpg_sign',
-                         'reedit_message', 'reuse_message', 'quiet'):
+                         'reedit_message', 'reuse_message', 'fixup', 'quiet'):
                     if v is None:
                         continue
                     if isinstance(v, bool):
@@ -136,17 +127,19 @@ class Ui:
 
             to_add = [f for f in files if os.path.exists(f)]
             if to_add:
-                util.system(['git', 'add', '-f', '-N', '--'] + to_add,
-                           onerr=util.Abort, errprefix=_("add failed"))
-            if opts['message'] is None:
-                util.system(['git', 'commit'] + args + ['--'] + list(files),
-                           onerr=util.Abort, errprefix=_("commit failed"))
+                system(['git', 'add', '-f', '-N', '--'] + to_add,
+                       onerr=Abort, errprefix=_("add failed"))
+            if not opts['message']:
+                system(['git', 'commit'] + args + ['--'] + list(files),
+                       onerr=Abort, errprefix=_("commit failed"))
             else:
-                util.system(['git', 'commit', '-F', name] + args + ['--'] + list(files),
-                           onerr=util.Abort, errprefix=_("commit failed"))
+                system(['git', 'commit', '-F', msgfile.name] + args + ['--'] + list(files),
+                       onerr=Abort, errprefix=_("commit failed"))
+            # refresh the index so that gitk doesn’t show empty staged diffs
+            systemcall(['git', 'update-index', '--ignore-submodules', '-q', '--ignore-missing', '--unmerged', '--refresh'])
 
         finally:
-            os.unlink(name)
+            msgfile.unlink(missing_ok=True)
 
 
 def main():
@@ -167,6 +160,7 @@ def main():
     parser.add_argument('-m', '--message', default=None, help='commit message')
     parser.add_argument('-c', '--reedit-message', metavar='COMMIT', default=None, help='reuse and edit message from specified commit')
     parser.add_argument('-C', '--reuse-message', metavar='COMMIT', default=None, help='reuse message from specified commit')
+    parser.add_argument('--fixup', metavar='COMMIT', default=None, help='create autosquash commit message to fixup specified commit')
     parser.add_argument('--reset-author', action='store_true', default=False, help='the commit is authored by me now (used with -C/-c/--amend)')
     parser.add_argument('-s', '--signoff', action='store_true', default=False, help='add Signed-off-by:')
     parser.add_argument('--amend', action='store_true', default=False, help='amend previous commit')
@@ -185,6 +179,9 @@ def main():
     opts = vars(args)
     opts['operation'] = subcommand
 
+    if opts['fixup'] and opts['fixup'].startswith('reword:'):
+        parser.error(_("Creating reword-only commits is not supported."))
+
     if subcommand == 'cstage':
         opts['index'] = True
 
@@ -198,8 +195,8 @@ def main():
     os.chdir(repo.path)
 
     try:
-        crecord_core.dorecord(ui, repo, None, **(opts))
-    except util.Abort as inst:
+        crecord_core.dorecord(ui, repo, **opts)
+    except Abort as inst:
         sinst = str(inst)
         if opts['quiet'] and 'commit failed' in sinst:
             sys.exit(5)
